
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

// content database "driver"


#include "base/polygraph.h"

#include "xstd/h/iostream.h"
#include <fstream.h>
#include "xstd/h/strstream.h"
#include "xstd/h/iomanip.h"

#include "xstd/gadgets.h"
#include "base/BStream.h"
#include "base/polyVersion.h"
#include "csm/ContentDbase.h"
#include "csm/cdbEntries.h"
#include "csm/XmlParser.h"


class CdbBuilder {
	public:
		CdbBuilder();
		virtual ~CdbBuilder();

		void db(ContentDbase *aDb);
		void configure(const String &aFname, const char *aBufB, const char *aBufE);
		virtual bool parse() = 0;

	protected:
		ContentDbase *theDb;
		String theFname;
		const char *theBufB;
		const char *theBufE;
};

class MarkupParser: public CdbBuilder {
	public:
		virtual bool parse();
};

class VerbatimParser: public CdbBuilder {
	public:
		virtual bool parse();
};


static String ThePrgName = "";
static enum { fmtVerbatim, fmtMarkup } TheFormat = fmtMarkup;


/* CdbBuilder */

CdbBuilder::CdbBuilder(): theDb(0), theBufB(0), theBufE(0) {
}

CdbBuilder::~CdbBuilder() {
}

void CdbBuilder::db(ContentDbase *aDb) {
	Assert(!theDb && aDb);
	theDb = aDb;
}

void CdbBuilder::configure(const String &aFname, const char *aBufB, const char *aBufE) {
	theFname = aFname;
	theBufB = aBufB;
	theBufE = aBufE;
}


/* MarkupParser */

bool MarkupParser::parse() {
	static XmlParser parser;
	(void)parser.parse(theBufB, theBufE);

	for (int i = 0; i < parser.nodeCount(); ++i) {
		const XmlParser::Node &node = parser.node(i);
		CdbEntry *e = 0;
		switch (node.type) {
			case XmlParser::Node::tpText:
				e = new CdbeText;
				break;

			case XmlParser::Node::tpTag:
				e = new CdbeTag;
				break;

			case XmlParser::Node::tpComment:
				e = new CdbeComment;
				break;

			default:
				Assert(false);
		}
		Assert(e);
		e->image(String(node.imageBeg, node.imageLen));
		theDb->add(e);
	}

	if (!parser.nodeCount())
		cerr << theFname << ": warning: no valid markup found" << endl;

	if (parser.tail().imageLen) {
		cerr << theFname << ": warning: ignoring markup leftovers starting at ";
		cerr.write(parser.tail().imageBeg, Min(parser.tail().imageLen, 40));
		cerr << endl;
	}

	return true;
}


/* VerbatimParser */

bool VerbatimParser::parse() {
	if (theBufB < theBufE) {
		CdbEntry *e = new CdbeBlob;
		e->image(String(theBufB, theBufE-theBufB));
		theDb->add(e);
		return true;
	}
	cerr << theFname << ": warning: empty verbatim file" << endl;
	return true;
}


static
bool readCont(ContentDbase *cdb, const String &fname) {
	const String hname = fname == "-" ? String("stdin") : fname;
	clog << ThePrgName << ": adding "
		<< (TheFormat == fmtMarkup ? "objects from " : "entire ")
		<< "'" << hname << "' ... ";

	const int savedCount = cdb->count();
	CdbBuilder *b = TheFormat == fmtMarkup ?
		(CdbBuilder*) new MarkupParser : (CdbBuilder*) new VerbatimParser;

	istream *is = fname == "-" ? 
		(istream*)&cin : (istream*)new ifstream(fname.cstr());
	b->db(cdb);

	ostrstream buf;
	char c;
	while (is->get(c))
		buf << c;
	const char *str = buf.str();
	b->configure(fname, str, str + buf.pcount());
	const bool res = b->parse();
	if (is != &cin)
		delete is;

	delete b;
	ostrstreamFreeze(buf, false);

	if (res)
		clog << "\t " << setw(6) << (cdb->count() - savedCount) << " object(s)";
	clog << endl;
	return res;
}

static
int doShow(const String &dbName) {
	ContentDbase *cdb = new ContentDbase;
	ifstream f(dbName.cstr());
	IBStream is;
	is.configure(&f, dbName);
	cdb->load(is);
	if (is.good()) {
		cdb->print(cout);
		return 0;
	}

	cerr << ThePrgName << ": error loading db from `" << dbName << "'" << endl;
	return -2;
}

static
int doAdd(const String &dbName, int fcount, char *fnames[]) {
	ContentDbase *cdb = new ContentDbase;
	fstream f(dbName.cstr(), ios::in|ios::out);

	{
		clog << ThePrgName << ": loading db from " << dbName << " ... ";
		IBStream is;
		is.configure(&f, dbName);
		cdb->load(is);
		clog << "\t " << setw(6) << cdb->count() << " objects" << endl;
	}

	if (fcount) {
		for (int i = 0; i < fcount; ++i)
			Must(readCont(cdb, fnames[i]));
	} else {
		Must(readCont(cdb, "-"));
	}

	{
		clog << ThePrgName << ": storing db in " << dbName << " ... ";
		f.clear();
		f.seekg(0);
		OBStream os;
		os.configure(&f, dbName);
		cdb->store(os);
		if (os.good())
			clog << "\t " << setw(6) << cdb->count() << " objects" << endl;
		else
			clog << "error: " << Error::Last() << endl;
	}
	f.close();
	return 0;
}

static
int usage() {
	(void)PolyVersion();
	cerr << "usage: " << ThePrgName << " <database.cdb> <command> [--option] [filename ...]" << endl
		<< "commands: " << endl
		<< "\tshow - " << " dump database contents to stdout" << endl
		<< "\tadd  - " << " add file(s) contents to the database" << endl
		<< "options: " << endl
		<< "\tformat <markup|verbatim> " << "split each file into markup tags or add as is" << endl
		;
	return 0;
}

int main(int argc, char *argv[]) {
	ThePrgName = argv[0];

	if (argc < 3)
		return usage();

	int argOff = 1;
	const String dbName = argv[argOff++];
	const String cmd = argv[argOff++];

	if (cmd != "show" && cmd != "add") {
		usage();
		cerr << "unknown command `" << cmd << "', expected `show' or `add'" << endl;
		return -1;
	}

	if (argOff < argc && strstr(argv[argOff], "--") == argv[argOff]) {
		const String option = argv[argOff++];
		if (option != "--format") {
			usage();
			cerr << "unknown option '" << option << "', expected 'format'" << endl;
			return -1;
		}
			
		const String format = argv[argOff++];
		if (format == "markup")
			TheFormat = fmtMarkup;
		else
		if (format == "verbatim")
			TheFormat = fmtVerbatim;
		else {
			usage();
			cerr << "unknown format '" << format << "', expected 'markup' or 'verbatim'" << endl;
			return -1;
		}
	}

	if (cmd == "show")
		return doShow(dbName);
	else
	if (cmd == "add")
		return doAdd(dbName, argc-argOff, argv+argOff);

	usage();
	cerr << "unknown command `" << cmd << "', expected `show' or `add'" << endl;
	return -1;
}
